
#include "main.h"
#include "Calculus.h"
#include "Geometry.h"
#include "Arithmetic.h"
#include "Stack.h"
#include "Convolve.h"
#include "Filter.h"
#include "File.h"
#include <sstream>

void Gradient::help() {
    printf("\n-gradient takes the backward differences in the dimension specified by the\n"
	   "argument. Values outside the image are assumed to be zero, so the first row,\n"
	   "or column, or frame, will not change, effectively storing the initial value\n"
	   "to make later integration easy. Multiple arguments can be given to differentiate\n"
	   "with respect to multiple dimensions in order (although the order does not matter).\n\n"
	   "Warning: Don't expect to differentiate more than twice and be able to get back\n"
	   "the image by integrating. Numerical errors will dominate.\n\n"
	   "Usage: ImageStack -load a.tga -gradient x y -save out.tga\n\n");
}

void Gradient::parse(vector<string> args) {
    assert(args.size() > 0, "-gradient requires at least one argument\n");
    for (size_t i = 0; i < args.size(); i++) {
	apply(stack(0), args[i]);
    }
}

// gradient can be called as gradient('t') or gradient("xyt") 
void Gradient::apply(Window im, string dimensions) {
    for (size_t i = 0; i < dimensions.size(); i++) {
	apply(im, dimensions[i]);
    }
}

void Gradient::apply(Window im, char dimension) {
    int mint = 0, minx = 0, miny = 0;
    int dt = 0, dx = 0, dy = 0;

    if (dimension == 'x') {
	dx = 1;
	minx = 1;
    } else if (dimension == 'y') {
	dy = 1;
	miny = 1;
    } else if (dimension == 't') {
	dt = 1;
	mint = 1;
    } else {
	panic("Must differentiate with respect to x, y, or t\n");
    }

    // walk backwards through the data, looking at the untouched data for the differences
    for (int t = im.frames - 1; t >= mint; t--) {
	for (int y = im.height - 1; y >= miny; y--) {
	    for (int x = im.width - 1; x >= minx; x--) {
		for (int c = 0; c < im.channels; c++) {
		    im(t, x, y)[c] -= im(t - dt, x - dx, y - dy)[c];
		}
	    }
	}
    }
}


void Integrate::help() {
    printf("\n-integrate computes partial sums along the given dimension. It is the\n"
	   "of the -gradient operator. Multiply dimensions can be given as arguments,\n"
	   "for example -integrate x y will produce a summed area table of an image.\n"
	   "Allowed dimensions are x, y, or t.\n\n"
	   "Warning: Don't expect to integrate more than twice and be able to get back\n"
	   "the image by differentiating. Numerical errors will dominate.\n\n"
	   "Usage: ImageStack -load a.tga -gradient x y -integrate y x -save a.tga\n\n");
}

void Integrate::parse(vector<string> args) {
    assert(args.size() > 0, "-integrate requires at least one argument\n");
    for (size_t i = 0; i < args.size(); i++) {
	apply(stack(0), args[i]);
    }	
}

// integrate can be called as integrate('t') or integrate("xyt") 
void Integrate::apply(Window im, string dimensions) {
    for (size_t i = 0; i < dimensions.size(); i++) {
	apply(im, dimensions[i]);
    }	
}

void Integrate::apply(Window im, char dimension) {
    int minx = 0, miny = 0, mint = 0;
    int dx = 0, dy = 0, dt = 0;

    if (dimension == 'x') {
	dx = 1;
	minx = 1;
    } else if (dimension == 'y') {
	dy = 1;
	miny = 1;
    } else if (dimension == 't') {
	dt = 1;
	mint = 1;
    } else {
	panic("Must integrate with respect to x, y, or t\n");
    }

    // walk forwards through the data, adding up as we go
    for (int t = mint; t < im.frames; t++) {
	for (int y = miny; y < im.height; y++) {
	    for (int x = minx; x < im.width; x++) {
		for (int c = 0; c < im.channels; c++) {
		    im(t, x, y)[c] += im(t - dt, x - dx, y - dy)[c];
		}
	    }
	}
    }

}


void Laplacian::help() {
    printf("\n-laplacian computes the gradient magnitude at each pixel in x and y. Temporal\n"
	   "gradients are ignored. The gradient is estimated using backward differences, and\n"
	   "the image is assumed to be zero outside its bounds.\n\n"
	   "Usage: ImageStack -load input.jpg -laplacian -save out.jpg\n");

}

void Laplacian::parse(vector<string> args) {
    assert(args.size() == 0, "-laplacian takes no arguments\n");
    Image im = apply(stack(0));
    pop();
    push(im);
}

Image Laplacian::apply(Window im) {
    Image out(im);
    Gradient::apply(im, 'x');
    Gradient::apply(out, 'y');

    for (int t = 0; t < im.frames; t++) {
	for (int y = 0; y < im.height; y++) {
	    for (int x = 0; x < im.width; x++) {
		for (int c = 0; c < im.channels; c++) {
		    out(t, x, y)[c] = im(t, x, y)[c] * im(t, x, y)[c] + out(t, x, y)[c] * out(t, x, y)[c];
		}
	    }
	}
    }

    return out;
}


void Poisson::help() {
    printf("-poisson assumes the stack contains gradients in x, y, and possibly t, and\n"
	   "attempts to find the image which fits those gradients best in a least squares\n"
	   "sense. It uses a conjugate gradient descent method. It can be called with zero\n"
	   "to two arguments. With two arguments, the first is the number of dimensions, 2,\n"
	   "or 3, and the second indicates how many seconds to run for. The arguments\n"
	   "default to 2 and 60 when one or zero arguments are given.\n\n"
	   "Usage: ImageStack -load dx.tmp dy.tmp \n"
	   "                  -load dt.tmp -poisson 3 30 -save out.tga\n\n");
}

void Poisson::parse(vector<string> args) {
    assert(args.size() < 3, "-poisson requires two or fewer arguments\n");
    int dimensions = 2;
    float seconds = 60;
    if (args.size() > 0) {
	dimensions = readInt(args[0]);
    }
    if (args.size() > 1) {
	seconds = readFloat(args[1]);
    }

    if (dimensions == 2) {
	push(apply(stack(1), stack(0), seconds));
    } else if (dimensions == 3) {
	push(apply(stack(2), stack(1), stack(0), seconds));
    } else {
	panic("number of dimensions must be two or three\n");
    }
	
}

Image Poisson::apply(Window dx, Window dy, float seconds) {
    assert(dx.width  == dy.width &&
	   dx.height == dy.height && 
	   dx.frames == dy.frames &&
	   dx.channels == dy.channels,
	   "derivatives must be matching size and number of channels\n");      
    
    float finishTime = currentTime() + seconds;


    bool recurse = false;
    Image estimate;

    // solve at a coarser scale
    if (dx.width * dx.height * dx.frames > 100) {
	recurse = true;
	int newWidth = (dx.width+1)/2;
	int newHeight = (dx.height+1)/2;
	int newFrames = (dx.frames+1)/2;
	Image smalldx = Resample::apply(dx, newFrames, newWidth, newHeight);
	Image smalldy = Resample::apply(dy, newFrames, newWidth, newHeight);
	Scale::apply(smalldx, 2);
	Scale::apply(smalldy, 2);
	Image smallSolution = Poisson::apply(smalldx, smalldy, seconds/3);
	estimate = Resample::apply(smallSolution, dx.frames, dx.width, dx.height);	
    } else {
	// start the estimate as black
	estimate = Image(dx.frames, dx.width, dx.height, dx.channels);
    }

    // figure out the sum of second derivatives
    Image newdx = Crop::apply(dx, 1, 0, dx.width, dx.height);
    Gradient::apply(newdx, 'x');
    Image newdy = Crop::apply(dy, 0, 1, dy.width, dy.height);
    Gradient::apply(newdy, 'y');
    Add::apply(newdx, newdy);
    Image b = newdx;

    float filter_[] = {0, 1, 0,
                       1, -4, 1,
                       0, 1, 0};
    Image filter(1, 3, 3, 1, filter_);

    Image solution = DeconvolveCG::apply(b, filter, finishTime - currentTime(), estimate);

    
    // reinstate the low freqs from the lower level solution
    if (0 && recurse) {
	Image blurry = Resample::apply(solution, estimate.frames, estimate.width, estimate.height);
	blurry = Resample::apply(blurry, solution.frames, solution.width, solution.height);
	Scale::apply(estimate, 0.5);
	Subtract::apply(solution, blurry);
	Add::apply(solution, estimate);
    }
    

    // fix the DC term
    vector<float> offset(solution.channels);
    for (int c = 0; c < solution.channels; c++) {
	offset[c] = (dx(0, 0, 0)[c] + dy(0, 0, 0)[c])/2 - solution(0, 0, 0)[c];
    }

    Offset::apply(solution, offset);

    return solution;

}

Image Poisson::apply(Window dt, Window dx, Window dy, float seconds) {
    assert(dx.width  == dy.width && dx.width == dt.width &&
	   dx.height == dy.height && dx.height == dt.height &&
	   dx.frames == dy.frames && dx.frames == dt.frames &&
	   dx.channels == dy.channels && dx.channels == dt.channels,
	   "derivatives must be matching size and number of channels\n");    
    Image out;
    return out;
}

/*

Image Poisson::apply(Window dx, Window dy, float termination) {
    assert(termination > 0 && termination < 1, 
	   "termination condition must be strictly between zero and one\n");
    assert(dx.width  == dy.width &&
	   dx.height == dy.height && 
	   dx.frames == dy.frames &&
	   dx.channels == dy.channels,
	   "derivatives must be matching size and number of channels\n");    

    int frames = dx.frames;
    int width = dx.width;
    int height = dx.height;
    int channels = dx.channels;

    Image solution;

    // if the image is large, solve it for a smaller version 
    if (width > 16 || height > 16) { // recursive case
	int lowerWidth = width < 16 ? width : (width + 1)/2;
	int lowerHeight = height < 16 ? height : (height + 1)/2;
	Image smallDx = Resample::apply(dx, lowerWidth, lowerHeight);
	Image smallDy = Resample::apply(dy, lowerWidth, lowerHeight);
	float scaleFactor = (float)(height * width) / (lowerHeight * lowerWidth);
	solution = apply(smallDx, smallDy, termination/10);
	Scale::apply(solution, scaleFactor);
	solution = Resample::apply(solution, width, height);
    } else {
	solution = Image(frames, width, height, channels);
    }    

    fprintf(stdout, "Solving at size %i %i\n", width, height);
    
    float stepSize = 0.5;
    double error, oldError = -1;
    int i = 0;
    for (;; i++) {
	error = 0;
 
	for (int t = 0; t < frames; t++) {
	    for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
		    for (int c = 0; c < channels; c++) {
			if (x > 0) {
			    // compare the known dx to the solution's dx here
			    float delta = dx(t, x, y)[c] - (solution(t, x, y)[c] - solution(t, x-1, y)[c]);
			    error += delta * delta;
			    delta *= stepSize;
			    solution(t, x, y)[c] += delta;
			    solution(t, x-1, y)[c] -= delta;
			    if (solution(t, x, y)[c] < 0) solution(t, x, y)[c] = 0;
			    if (solution(t, x-1, y)[c] < 0) solution(t, x-1, y)[c] = 0;
			}		
			    
			if (y > 0) {
			    // compare the known dy to the solution's dy here
			    float delta = dy(t, x, y)[c] - (solution(t, x, y)[c] - solution(t, x, y-1)[c]);
			    error += delta * delta;
			    delta *= stepSize;
			    solution(t, x, y)[c] += delta;
			    solution(t, x, y-1)[c] -= delta;
			    if (solution(t, x, y)[c] < 0) solution(t, x, y)[c] = 0;
			    if (solution(t, x, y-1)[c] < 0) solution(t, x, y-1)[c] = 0;
			}
		    }
		}
	    }
	}

	for (int t = frames - 1; t >= 0; t--) {
	    for (int y = height - 1; y >= 0; y--) {
		for (int x = width - 1; x >= 0; x--) {
		    for (int c = 0; c < channels; c++) {
			if (x > 0) {
			    // compare the known dx to the solution's dx here
			    float delta = dx(t, x, y)[c] - (solution(t, x, y)[c] - solution(t, x-1, y)[c]);
			    error += delta * delta;
			    delta *= stepSize;
			    solution(t, x, y)[c] += delta;
			    solution(t, x-1, y)[c] -= delta;
			    if (solution(t, x, y)[c] < 0) solution(t, x, y)[c] = 0;
			    if (solution(t, x-1, y)[c] < 0) solution(t, x-1, y)[c] = 0;
			}		
			    
			if (y > 0) {
			    // compare the known dy to the solution's dy here
			    float delta = dy(t, x, y)[c] - (solution(t, x, y)[c] - solution(t, x, y-1)[c]);
			    error += delta * delta;
			    delta *= stepSize;
			    solution(t, x, y)[c] += delta;
			    solution(t, x, y-1)[c] -= delta;
			    if (solution(t, x, y)[c] < 0) solution(t, x, y)[c] = 0;
			    if (solution(t, x, y-1)[c] < 0) solution(t, x, y-1)[c] = 0;
			}		    
		    }
		}
	    }
	}

	error = sqrt(error) / (frames * width * height * channels);
	if (i % 100 == 0 && i > 0) fprintf(stdout, "Iteration %i\tRMS Error: %3.10f\n", i, error);
	if (oldError > 0 && error <= oldError && (oldError - error) < termination * oldError) break;
	if (i > 100000) break;
	oldError = error;
    }

    fprintf(stdout, "Iterations %i\tRMS Error: %3.10f\n", i, error);

    // this solver has a degree of freedom in the DC value,
    // so set the top left to the appropriate value
    vector<float> offset(channels);
    for (int c = 0; c < channels; c++) {
	offset[c] = (dx(0, 0, 0)[c] + dy(0, 0, 0)[c])/2 - solution(0, 0, 0)[c];
    }

    Offset::apply(solution, offset);
    return solution;
}

Image Poisson::apply(Window dt, Window dx, Window dy, float termination) {
    assert(termination > 0 && termination < 1, 
	   "termination condition must be strictly between zero and one\n");
    assert(dx.width  == dy.width && dx.width == dt.width &&
	   dx.height == dy.height && dx.height == dt.height &&
	   dx.frames == dy.frames && dx.frames == dt.frames &&
	   dx.channels == dy.channels && dx.channels == dt.channels,
	   "derivatives must be matching size and number of channels\n");    

    int frames = dx.frames;
    int width = dx.width;
    int height = dx.height;
    int channels = dx.channels;

    Image solution;

    // if the image is large, solve it for a smaller version 
    if (width > 16 || height > 16 || frames > 16) { // recursive case
	int lowerWidth = width < 16 ? width : (width+1)/2;
	int lowerHeight = height < 16 ? height : (height+1)/2;
	int lowerFrames = frames < 16 ? frames : (frames+1)/2;
	Image smallDt = Resample::apply(dt, lowerFrames, lowerWidth, lowerHeight);
	Image smallDx = Resample::apply(dx, lowerFrames, lowerWidth, lowerHeight);
	Image smallDy = Resample::apply(dy, lowerFrames, lowerWidth, lowerHeight);
	float scaleFactor = (float)(frames * height * width) / (lowerFrames * lowerHeight * lowerWidth);
	solution = apply(smallDt, smallDx, smallDy, termination/10);
	Scale::apply(solution, scaleFactor);
	solution = Resample::apply(solution, frames, width, height);
    } else {
	solution = Image(frames, width, height, channels);
    }    

    fprintf(stdout, "Solving at size %i %i %i\n", frames, width, height);

    float error, oldError = -1;
    int i = 0;
    for (;; i++) {
	error = 0;

	for (int t = 0; t < frames; t++) {
	    for (int y = 0; y < height; y++) {
		for (int x = 0; x < width; x++) {
		    for (int c = 0; c < channels; c++) {
			if (t > 0) {
			    // compare the known dy to the solution's dy here
			    float delta = dt(t, x, y)[c] - (solution(t, x, y)[c] - solution(t-1, x, y)[c]);
			    error += delta * delta;
			    delta *= 0.5;
			    solution(t, x, y)[c] -= delta;
			    solution(t-1, x, y)[c] += delta;
			}

			if (x > 0) {
			    // compare the known dx to the solution's dx here
			    float delta = dx(t, x, y)[c] - (solution(t, x, y)[c] - solution(t, x-1, y)[c]);
			    error += delta * delta;
			    delta *= 0.5;
			    solution(t, x, y)[c] -= delta;
			    solution(t, x-1, y)[c] += delta;
			}		
			    
			if (y > 0) {
			    // compare the known dy to the solution's dy here
			    float delta = dy(t, x, y)[c] - (solution(t, x, y)[c] - solution(t, x, y-1)[c]);
			    error += delta * delta;
			    delta *= 0.5;
			    solution(t, x, y)[c] -= delta;
			    solution(t, x, y-1)[c] += delta;
			}

		    }
		}
	    }
	}

	for (int t = frames - 1; t >= 0; t--) {
	    for (int y = height - 1; y >= 0; y--) {
		for (int x = width - 1; x >= 0; x--) {
		    for (int c = 0; c < channels; c++) {
			if (t > 0) {
			    // compare the known dy to the solution's dy here
			    float delta = dt(t, x, y)[c] - (solution(t, x, y)[c] - solution(t-1, x, y)[c]);
			    error += delta * delta;
			    delta *= 0.5;
			    solution(t, x, y)[c] -= delta;
			    solution(t-1, x, y)[c] += delta;
			}

			if (x > 0) {
			    // compare the known dx to the solution's dx here
			    float delta = dx(t, x, y)[c] - (solution(t, x, y)[c] - solution(t, x-1, y)[c]);
			    error += delta * delta;
			    delta *= 0.5;
			    solution(t, x, y)[c] -= delta;
			    solution(t, x-1, y)[c] += delta;
			}		
			    
			if (y > 0) {
			    // compare the known dy to the solution's dy here
			    float delta = dy(t, x, y)[c] - (solution(t, x, y)[c] - solution(t, x, y-1)[c]);
			    error += delta * delta;
			    delta *= 0.5;
			    solution(t, x, y)[c] -= delta;
			    solution(t, x, y-1)[c] += delta;
			}
		    }
		}
	    }
	}

	error = sqrt(error) / (width * height * frames * channels);
	if (i % 100 == 0 && i > 0) fprintf(stdout, "Iteration %i\tRMS Error: %3.10f\n", i, error);
	if (oldError > 0 && error <= oldError && (oldError - error) < termination * oldError) break;
	oldError = error;
    }

    fprintf(stdout, "Iterations %i\tRMS Error: %3.10f\n", i, error);

    // this solver has a degree of freedom in the DC value,
    // so set the top left to the appropriate value
    vector<float> offset(channels);
    for (int c = 0; c < channels; c++) {
	offset[c] = (dx(0, 0, 0)[c] + dy(0, 0, 0)[c])/2 - solution(0, 0, 0)[c];
    }

    Offset::apply(solution, offset);
    return solution;
}


*/
