#include "main.h"
#include "Filter.h"
#include "Convolve.h"
#include "Color.h"
#include "Geometry.h"
#include "Arithmetic.h"

void GaussianBlur::help() {
    printf("\n-gaussianblur takes a floating point frames, width, and height, and performs\n"
	   "a gaussian blur with those standard deviations. The blur is performed out to\n"
	   "three standard deviations. If given only two arguments, it performs a blur in\n"
	   "x and y only. If given one argument, it performs the blur in x and y with\n"
	   "filter width the same as height.\n\n"
	   "Usage: ImageStack -load in.jpg -gaussianblur 5 -save blurry.jpg\n\n");
}

void GaussianBlur::parse(vector<string> args) {
    float frames = 0, width = 0, height = 0;
    if (args.size() == 1) {
	width = height = readFloat(args[0]);
    } else if (args.size() == 2) {
	width = readFloat(args[0]);
	height = readFloat(args[1]);
    } else if (args.size() == 3) {
	frames = readFloat(args[0]);
	width  = readFloat(args[1]);
	height = readFloat(args[2]);
    } else {
	panic("-gaussianblur takes one, two, or three arguments\n");
    }

    Image im = apply(stack(0), frames, width, height);
    pop();
    push(im);
}

Image GaussianBlur::apply(Window im, float filterFrames, float filterWidth, float filterHeight) {
    Image out(im);

    if (filterFrames != 0) {
	// make the frames filter
	int size = (int)(filterFrames * 6 + 1) | 1;       
	// even tiny filters should do something, otherwise we
	// wouldn't have called this function.
	if (size == 1) size = 3;
	int radius = size / 2;
	Image filter(size, 1, 1, 1);
	float sum = 0;
	for (int i = 0; i < size; i++) {
            float diff = (i-radius)/filterFrames;
	    float value = expf(-diff * diff / 2);
	    filter(i, 0, 0)[0] = value;
	    sum += value;
	}
	
	for (int i = 0; i < size; i++) {
	    filter(i, 0, 0)[0] /= sum;
	}
	
	out = Convolve::apply(out, filter);
    }

    if (filterWidth != 0) {
	// make the width filter
	int size = (int)(filterWidth * 6 + 1) | 1;
	// even tiny filters should do something, otherwise we
	// wouldn't have called this function.
	if (size == 1) size = 3;
	int radius = size / 2;
	Image filter(1, size, 1, 1);
	float sum = 0;
	for (int i = 0; i < size; i++) {
            float diff = (i-radius)/filterWidth;
	    float value = expf(-diff * diff / 2);
	    filter(0, i, 0)[0] = value;
	    sum += value;
	}
	
	for (int i = 0; i < size; i++) {
	    filter(0, i, 0)[0] /= sum;
	}
	
	out = Convolve::apply(out, filter);
    }

    if (filterHeight != 0) {
	// make the height filter
	int size = (int)(filterHeight * 6 + 1) | 1;
	// even tiny filters should do something, otherwise we
	// wouldn't have called this function.
	if (size == 1) size = 3;
	int radius = size / 2;
	Image filter(1, 1, size, 1);
	float sum = 0;
	for (int i = 0; i < size; i++) {
            float diff = (i-radius)/filterHeight;
	    float value = expf(-diff * diff / 2);
	    filter(0, 0, i)[0] = value;
	    sum += value;
	}
	
	for (int i = 0; i < size; i++) {
	    filter(0, 0, i)[0] /= sum;
	}
	
	out = Convolve::apply(out, filter);
    }

    return out;
}

// This blur implementation was contributed by Tyler Mullen as a
// CS448F project. A competition was held, and this method was found
// to be much faster than other IIRs, filtering by resampling,
// iterated rect filters, and polynomial integral images. The method
// was modified by Andrew Adams to be more ImageStacky (i.e. use
// structures more idiomatic to ImageStack like pointer marching), to
// work for larger sized blurs, and to cover more unusual cases.

void FastBlur::help() {
    printf("\n-fastblur takes a floating point frames, width, and height, and performs\n"
	   "a fast approximate gaussian blur with those standard deviations using the IIR\n"
	   "method of van Vliet et al. If given only two arguments, it performs a blur in\n"
	   "x and y only. If given one argument, it performs the blur in x and y with\n"
	   "filter width the same as height.\n\n"
	   "Usage: ImageStack -load in.jpg -fastblur 5 -save blurry.jpg\n\n");
}

void FastBlur::parse(vector<string> args) {
    float frames = 0, width = 0, height = 0;
    if (args.size() == 1) {
	width = height = readFloat(args[0]);
    } else if (args.size() == 2) {
	width = readFloat(args[0]);
	height = readFloat(args[1]);
    } else if (args.size() == 3) {
	frames = readFloat(args[0]);
	width  = readFloat(args[1]);
	height = readFloat(args[2]);
    } else {
	panic("-fastblur takes one, two, or three arguments\n");
    }

    apply(stack(0), frames, width, height);
}

void FastBlur::apply(Window im, float filterFrames, float filterWidth, float filterHeight, bool addMargin) {
    assert(filterFrames >= 0 &&
	   filterWidth >= 0 &&
	   filterHeight >= 0,
	   "Filter sizes must be non-negative\n");

    // Prevent filtering in useless directions
    if (im.width == 1)  filterWidth = 0;
    if (im.height == 1) filterHeight = 0;
    if (im.frames == 1) filterFrames = 0;

    // Filter in very narrow directions using the regular Gaussian, as
    // the IIR requires a few pixels to get going. If the Gaussian
    // blur is very narrow, also revert to the naive method, as IIR
    // won't work.
    if (filterFrames > 0 && (im.frames < 16 || filterFrames < 0.5)) {
	Image blurry = GaussianBlur::apply(im, filterFrames, 0, 0);
	FastBlur::apply(blurry, 0, filterWidth, filterHeight);	
	Paste::apply(im, blurry, 0, 0, 0);
	return;
    }

    if (filterWidth > 0 && (im.width < 16 || filterWidth < 0.5)) {
	Image blurry = GaussianBlur::apply(im, 0, filterWidth, 0);
	FastBlur::apply(blurry, filterFrames, 0, filterHeight);	
	Paste::apply(im, blurry, 0, 0, 0);
	return;
    }

    if (filterHeight > 0 && (im.height < 16 || filterHeight < 0.5)) {
	Image blurry = GaussianBlur::apply(im, 0, 0, filterHeight);
	FastBlur::apply(blurry, filterFrames, filterWidth, 0);
	Paste::apply(im, blurry, 0, 0, 0);
	return;
    }

    // IIR filtering fails if the std dev is similar to the image
    // size, because it displays a bias towards the edge values on the
    // starting side. We solve this by adding a margin and using
    // homogeneous weights.
    if (addMargin && (im.frames / filterFrames < 8 ||
		      im.width / filterWidth < 8 ||
		      im.height / filterHeight < 8)) {

	int marginT = (int)(filterFrames);
	int marginX = (int)(filterWidth);
	int marginY = (int)(filterHeight);

	Image bigger(im.frames+2*marginT, im.width+2*marginX, im.height+2*marginY, im.channels+1);
	for (int t = 0; t < im.frames; t++) {
	    for (int y = 0; y < im.height; y++) {
		float *imPtr = im(t, 0, y);
		float *biggerPtr = bigger(t+marginT, marginX, y+marginY);
		for (int x = 0; x < im.width; x++) {
		    *biggerPtr++ = 1;
		    for (int c = 0; c < im.channels; c++) {
			*biggerPtr++ = *imPtr++;
		    }
		}
	    }
	}

	FastBlur::apply(bigger, filterFrames, filterWidth, filterHeight, false);
	for (int t = 0; t < im.frames; t++) {
	    for (int y = 0; y < im.height; y++) {
		float *imPtr = im(t, 0, y);
		float *biggerPtr = bigger(t+marginT, marginX, y+marginY);
		for (int x = 0; x < im.width; x++) {
		    float w = 1.0f/(*biggerPtr++);
		    for (int c = 0; c < im.channels; c++) {
			*imPtr++ = w*(*biggerPtr++);
		    }		    
		}
	    }
	}

	return;	
    }    

    // now perform the blur
    if (filterWidth > 32) {
	// for large filters, we decompose into a dense blur and a
	// sparse blur, by spacing out the taps on the IIR
	float remainingStdDev = sqrtf(filterWidth*filterWidth - 32*32);
	int tapSpacing = (int)(remainingStdDev / 32 + 1);	    
	blurX(im, remainingStdDev/tapSpacing, tapSpacing);
	blurX(im, 32, 1);
    } else if (filterWidth > 0) {
	blurX(im, filterWidth, 1);
    }

    if (filterHeight > 32) {
	float remainingStdDev = sqrtf(filterHeight*filterHeight - 32*32);
	int tapSpacing = (int)(remainingStdDev / 32 + 1);	    
	blurY(im, remainingStdDev/tapSpacing, tapSpacing);
	blurY(im, 32, 1);
    } else if (filterHeight > 0) {
	blurY(im, filterHeight, 1);
    }

    if (filterFrames > 32) {
	float remainingStdDev = sqrtf(filterFrames*filterFrames - 32*32);
	int tapSpacing = (int)(remainingStdDev / 32 + 1);	    
	blurT(im, remainingStdDev/tapSpacing, tapSpacing);
	blurT(im, 32, 1);
    } else if (filterFrames > 0) {
	blurT(im, filterFrames, 1);
    }
}

void FastBlur::blurX(Window im, float sigma, int ts) {
    if (sigma == 0) return;

    float *imPtr;
    // blur in the x-direction
    float c0, c1, c2, c3;
    calculateCoefficients(sigma, &c0, &c1, &c2, &c3);
    
    float invC01 = 1.0f/(c0+c1);
    float invC012 = 1.0f/(c0+c1+c2);
    
    // we step through each row of each frame, and apply a forwards and then
    // a backwards pass of our IIR filter to approximate Gaussian blurring
    // in the x-direction
    for (int t = 0; t < im.frames; t++) {
	for (int y = 0; y < im.height; y++) {
	    // forward pass
	    
	    // use a zero boundary condition in the homogeneous
	    // sense (ie zero weight outside the image, divide by
	    // the sum of the weights)
	    for (int j = 0; j < ts; j++) {
		for (int c = 0; c < im.channels; c++){
		    //im(t,0*ts+j,y)[c] = c0*im(t,0*ts+j,y)[c] / c0;
		    im(t,ts+j,y)[c] = (c0*im(t,ts+j,y)[c] + c1*im(t,j,y)[c]) * invC01;
		    im(t,2*ts+j,y)[c] = (c0*im(t,2*ts+j,y)[c] + c1*im(t,ts+j,y)[c] + c2*im(t,j,y)[c]) * invC012;
		}
	    }
	    
	    
	    // now apply the forward filter
	    imPtr = im(t, 3*ts, y);
	    int lookBack = -im.xstride*ts;
	    for (int i = 0; i < (im.width-3*ts)*im.channels; i++) {
		// our forwards IIR equation
		imPtr[0] = (c0*imPtr[0] +
			    c1*imPtr[lookBack] +
			    c2*imPtr[lookBack*2] + 
			    c3*imPtr[lookBack*3]);
		imPtr++;
	    }
	    
	    // use a zero boundary condition in the homogeneous
	    // sense	       
	    int x = im.width-3*ts;
	    for (int j = 0; j < ts; j++) {
		for (int c = 0; c < im.channels; c++){
		    //im(t,x+2*ts+j,y)[c] = c0*im(t,x+2*ts+j,y)[c] / c0;
		    im(t,x+ts+j,y)[c] = (c0*im(t,x+ts+j,y)[c] + c1*im(t,x+2*ts+j,y)[c]) * invC01;
		    im(t,x+j,y)[c] = (c0*im(t,x+j,y)[c] + c1*im(t,x+ts+j,y)[c] + c2*im(t,x+2*ts+j,y)[c]) * invC012;
		}				
	    }
	    
	    // backward pass
	    imPtr = im(t, im.width-3*ts, y);
	    lookBack = im.xstride*ts;
	    for (int i = 0; i < (im.width-3*ts)*im.channels; i++) {
		imPtr--;
		imPtr[0] = (c0*imPtr[0] +
			    c1*imPtr[lookBack] +
			    c2*imPtr[lookBack*2] + 
			    c3*imPtr[lookBack*3]);
	    }
	}
    } 
}

void FastBlur::blurY(Window im, float sigma, int ts) {
    if (sigma == 0) return;

    float c0, c1, c2, c3;
    calculateCoefficients(sigma, &c0, &c1, &c2, &c3);
    float invC01 = 1.0f/(c0+c1);
    float invC012 = 1.0f/(c0+c1+c2);

    float *imPtr;

    // blur in the y-direction
    //  we do the same thing here as in the x-direction
    //  but we apply im.width different filters in parallel,
    //  for cache coherency's sake, first all going in the "forwards"
    //  direction, and then all going in the "backwards" direction
    for (int t = 0; t < im.frames; t++) {
	// use a zero boundary condition in the homogeneous
	// sense (ie zero weight outside the image, divide by
	// the sum of the weights)
	for (int j = 0; j < ts; j++) {
	    for (int x = 0; x < im.width; x++) {
		for (int c = 0; c < im.channels; c++) {		    
		    //im(t,x,j)[c] = c0*im(t,x,j)[c] / c0;
		    im(t,x,ts+j)[c] = (c0*im(t,x,ts+j)[c] + c1*im(t,x,j)[c]) * invC01;
		    im(t,x,2*ts+j)[c] = (c0*im(t,x,2*ts+j)[c] + c1*im(t,x,ts+j)[c] + c2*im(t,x,j)[c]) * invC012;
		}
	    }
	}
	
	// forward pass
	
	int lookBack = -im.ystride*ts;
	for (int y = 3*ts; y < im.height; y++) {
	    imPtr = im(t, 0, y);
	    for (int i = 0; i < im.width*im.channels; i++) {
		imPtr[0] = (c0*imPtr[0] +
			    c1*imPtr[lookBack] +
			    c2*imPtr[lookBack*2] + 
			    c3*imPtr[lookBack*3]);
		imPtr++;
	    }
	}
	
	// use a zero boundary condition in the homogeneous
	// sense (ie zero weight outside the image, divide by
	// the sum of the weights)
	int y = im.height-3*ts;
	for (int j = 0; j < ts; j++) {	       
	    for (int x = 0; x < im.width; x++) {
		for (int c = 0; c < im.channels; c++) {		    
		    //im(t,x,y+ts*2+j)[c] = c0*im(t,x,y+ts*2+j)[c] / c0;
		    im(t,x,y+ts+j)[c] = (c0*im(t,x,y+ts+j)[c] + c1*im(t,x,y+ts*2+j)[c]) * invC01;
		    im(t,x,y+j)[c] = (c0*im(t,x,y+j)[c] + c1*im(t,x,y+ts+j)[c] + c2*im(t,x,y+ts*2+j)[c]) * invC012;
		}
	    }
	}
	
	// backward pass	  
	lookBack = im.ystride*ts;
	for (int y = im.height-3*ts-1; y >= 0; y--) {
	    imPtr = im(t, 0, y);
	    for (int i = 0; i < im.width*im.channels; i++) {
		imPtr[0] = (c0*imPtr[0] +
			    c1*imPtr[lookBack] +
			    c2*imPtr[lookBack*2] + 
			    c3*imPtr[lookBack*3]);
		imPtr++;
	    }
	}
    }	    
}

void FastBlur::blurT(Window im, float sigma, int ts) {
    if (sigma == 0) return;

    float c0, c1, c2, c3;
    calculateCoefficients(sigma, &c0, &c1, &c2, &c3);
    float invC01 = 1.0f/(c0+c1);
    float invC012 = 1.0f/(c0+c1+c2);

    float *imPtr;
    
    // blur in the t-direction 
    // this is the same strategy as blurring in y, but we swap t
    // for y everywhere
    for (int y = 0; y < im.height; y++) {
	// use a zero boundary condition in the homogeneous
	// sense (ie zero weight outside the image, divide by
	// the sum of the weights)
	for (int j = 0; j < ts; j++) {
	    for (int x = 0; x < im.width; x++) {
		for (int c = 0; c < im.channels; c++) {		    
		    //im(j,x,y)[c] = c0*im(j,x,y)[c] / c0;
		    im(ts+j,x,y)[c] = (c0*im(ts+j,x,y)[c] + c1*im(j,x,y)[c]) * invC01;
		    im(2*ts+j,x,y)[c] = (c0*im(2*ts+j,x,y)[c] + c1*im(ts+j,x,y)[c] + c2*im(j,x,y)[c]) * invC012;
		}
	    }
	}
	
	// forward pass
	
	int lookBack = -im.tstride*ts;
	for (int t = 3*ts; t < im.frames; t++) {
	    imPtr = im(t, 0, y);
	    for (int i = 0; i < im.width*im.channels; i++) {
		imPtr[0] = (c0*imPtr[0] +
			    c1*imPtr[lookBack] +
			    c2*imPtr[lookBack*2] + 
			    c3*imPtr[lookBack*3]);
		imPtr++;
	    }
	}
	
	// use a zero boundary condition in the homogeneous
	// sense (ie zero weight outside the image, divide by
	// the sum of the weights)
	int t = im.frames-3*ts;
	for (int j = 0; j < ts; j++) {
	    for (int x = 0; x < im.width; x++) {
		for (int c = 0; c < im.channels; c++) {		    
		    //im(t+2*ts+j,x,y)[c] = c0*im(t+2*ts+j,x,y)[c] / c0;
		    im(t+ts+j,x,y)[c] = (c0*im(t+ts+j,x,y)[c] + c1*im(t+2*ts+j,x,y)[c]) * invC01;
		    im(t+j,x,y)[c] = (c0*im(t+j,x,y)[c] + c1*im(t+ts+j,x,y)[c] + c2*im(t+2*ts+j,x,y)[c]) * invC012;
		}
	    }
	}
	
	// backward pass	  
	lookBack = im.tstride*ts;
	for (int t = im.frames-3*ts-1; t >= 0; t--) {
	    imPtr = im(t, 0, y);
	    for (int i = 0; i < im.width*im.channels; i++) {
		imPtr[0] = (c0*imPtr[0] +
			    c1*imPtr[lookBack] +
			    c2*imPtr[lookBack*2] + 
			    c3*imPtr[lookBack*3]);
		imPtr++;
	    }
	}	    
    }
}


void FastBlur::calculateCoefficients(float sigma, float *c0, float *c1, float *c2, float *c3) {
    // performs the necessary conversion between the sigma of a Gaussian blur
    // and the coefficients used in the IIR filter
    
    float q;

    assert(sigma >= 0.5, "To use IIR filtering, standard deviation of blur must be >= 0.5\n");

    if (sigma < 2.5) {
	q = (3.97156 - 4.14554*sqrtf(1 - 0.26891*sigma));
    } else {
	q = 0.98711*sigma - 0.96330;
    }
    
    float denom = 1.57825 + 2.44413*q + 1.4281*q*q + 0.422205*q*q*q;
    *c1 = (2.44413*q + 2.85619*q*q + 1.26661*q*q*q)/denom;
    *c2 = -(1.4281*q*q + 1.26661*q*q*q)/denom;
    *c3 = (0.422205*q*q*q)/denom;
    *c0 = 1 - (*c1 + *c2 + *c3);   
}

void LanczosBlur::help() {
    printf("\n-lanczosblur convolves the current image by a three lobed lanczos filter. A\n"
	   "lanczos filter is a kind of windowed sinc. One argument is interpreted as filter\n"
	   "width (= filter height). Filter frames is assumed to be one. Two arguments are\n"
	   "interpreted as filter width and filter height. Three arguments are interpreted\n"
	   "as filter frames, filter width, and filter height. The arguments may all be\n"
	   "arbitrary positive real numbers.\n\n"
	   "Usage: ImageStack -load big.jpg -lanczosblur 2 -subsample 2 2 0 0 -save small.jpg\n\n");

}

void LanczosBlur::parse(vector<string> args) {
    float frames = 0, width = 0, height = 0;
    if (args.size() == 1) {
	width = height = readFloat(args[0]);
    } else if (args.size() == 2) {
	width = readFloat(args[0]);
	height = readFloat(args[1]);
    } else if (args.size() == 3) {
	frames = readFloat(args[0]);
	width  = readFloat(args[1]);
	height = readFloat(args[2]);
    } else {
	panic("-lanczosblur takes one, two, or three arguments\n");
    }

    Image im = apply(stack(0), frames, width, height);
    pop();
    push(im);
}

Image LanczosBlur::apply(Window im, float filterFrames, float filterWidth, float filterHeight) {
    Image out(im);

    if (filterFrames != 0) {
	// make the frames filter
	int size = (int)(filterFrames * 6 + 1) | 1;
	int radius = size / 2;
	Image filter(size, 1, 1, 1);
	float sum = 0;
	for (int i = 0; i < size; i++) {
	    float value = lanczos_3((i-radius) / filterFrames);
	    filter(i, 0, 0)[0] = value;
	    sum += value;
	}
	
	for (int i = 0; i < size; i++) {
	    filter(i, 0, 0)[0] /= sum;
	}
	
	out = Convolve::apply(out, filter);
    }

    if (filterWidth != 0) {
	// make the width filter
	int size = (int)(filterWidth * 6 + 1) | 1;
	int radius = size / 2;
	Image filter(1, size, 1, 1);
	float sum = 0;
	for (int i = 0; i < size; i++) {
	    float value = lanczos_3((i-radius) / filterWidth);
	    filter(0, i, 0)[0] = value;
	    sum += value;
	}
	
	for (int i = 0; i < size; i++) {
	    filter(0, i, 0)[0] /= sum;
	}
	
	out = Convolve::apply(out, filter);
    }

    if (filterHeight != 0) {
	// make the height filter
	int size = (int)(filterHeight * 6 + 1) | 1;
	int radius = size / 2;
	Image filter(1, 1, size, 1);
	float sum = 0;
	for (int i = 0; i < size; i++) {
	    float value = lanczos_3((i-radius) / filterHeight);
	    filter(0, 0, i)[0] = value;
	    sum += value;
	}
	
	for (int i = 0; i < size; i++) {
	    filter(0, 0, i)[0] /= sum;
	}
	
	out = Convolve::apply(out, filter);
    }

    return out;

}

void Bilateral::help() {
    printf("\n-bilateral performs a bilateral filter on the image. Using a bilateral filter,\n"
           "as in a gaussian blur, each pixel is replaced by a gaussian weighted combination\n"
           "of its neighbours. The distance function, however, measures distance in space and\n"
           "color. The two arguments to the filter are the spatial standard deviation and the\n"
           "color space standard deviation.\n\n"
           "Usage: ImageStack -load in.jpg -bilateral 1 0.4 -save out.jpg\n\n");
}

void Bilateral::parse(vector<string> args) {
    assert(args.size() == 2, "-bilateral takes two arguments\n");
    Image im = apply(stack(0), readFloat(args[0]), readFloat(args[1]));
    pop();
    push(im);    
}

Image Bilateral::apply(Window im, float spatialSigma, float colorSigma) {

    Image out(im.frames, im.width, im.height, im.channels);

    // make the spatial filter
    int filterSizeT = (int)(spatialSigma * 8 + 1) | 1;
    int filterSizeX = (int)(spatialSigma * 8 + 1) | 1;
    int filterSizeY = (int)(spatialSigma * 8 + 1) | 1;

    // common case optimization
    if (im.frames == 1) filterSizeT = 1;

    Image filter(filterSizeT, filterSizeX, filterSizeY, 1);

    for (int t = 0; t < filter.frames; t++) {
	for (int y = 0; y < filter.height; y++) {
	    for (int x = 0; x < filter.width; x++) {		
		float dt = (t - filter.frames / 2) / spatialSigma;
		float dx = (x - filter.width / 2) / spatialSigma;
		float dy = (y - filter.height / 2) / spatialSigma;
		float distance = dt * dt + dx * dx + dy * dy;
		float value = expf(-distance/2);
		filter(t, x, y)[0] = value;
	    }
	}
    }

    // convolve
    int xoff = (filter.width - 1)/2;
    int yoff = (filter.height - 1)/2;
    int toff = (filter.frames - 1)/2;

    double colorSigmaMult = 1.0/colorSigma;

    for (int t = 0; t < im.frames; t++) {
        for (int y = 0; y < im.height; y++) {
            for (int x = 0; x < im.width; x++) {
                double totalWeight = 0;
                for (int dt = -toff; dt <= toff; dt++) {		    
                    int imt = t + dt;
		    if (imt < 0) continue;
		    if (imt >= im.frames) break;
                    int filtert = dt + toff;
                    for (int dy = -yoff; dy <= yoff; dy++) {
			int imy = y + dy;
			if (imy < 0) continue;
			if (imy >= im.height) break;
                        int filtery = dy + yoff;
                        for (int dx = -xoff; dx <= xoff; dx++) {
			    int imx = x + dx;
			    if (imx < 0) continue;
			    if (imx >= im.width) break;
                            int filterx = dx + xoff;
                            float weight = filter(filtert, filterx, filtery)[0];
                            float colorDistance = 0;                           
                            for (int c = 0; c < im.channels; c++) {
                                float diff = (im(imt, imx, imy)[c] - im(t, x, y)[c]) * colorSigmaMult;
                                colorDistance += diff * diff;
                            }
			    weight *= fastexp(-colorDistance/2);
			    if (weight <= 0.006738) weight = 0;
                            totalWeight += weight;
                            for (int c = 0; c < im.channels; c++) {
                                out(t, x, y)[c] += weight * im(imt, imx, imy)[c];
                            }
                        }
                    }
                }
                for (int c = 0; c < im.channels; c++) {
                    out(t, x, y)[c] /= totalWeight;
                }
            }
        }
    }
    

    return out;
}


void JointBilateral::help() {
    printf("\n-jointbilateral performs a joint bilateral filter on the image. The top image,\n"
	   "on the stack is blurred, using the second image as a reference for which edges\n"
	   "to not blur across. See -help bilateral for details on the bilateral filter.\n\n"
           "Usage: ImageStack -load ref.jpg -load in.jpg -jointbilateral 1 0.4 -save out.jpg\n\n");
}

void JointBilateral::parse(vector<string> args) {
    assert(args.size() == 2, "-jointbilateral takes two arguments\n");
    Image im = apply(stack(0), stack(1), readFloat(args[0]), readFloat(args[1]));
    pop();
    push(im);    
}

Image JointBilateral::apply(Window im, Window ref, float spatialSigma, float colorSigma) {

    assert(im.width == ref.width && im.height == ref.height && im.frames == ref.frames, 
	   "The reference image must be the same size as the image to be filtered.\n");

    Image out(im.frames, im.width, im.height, im.channels);

    // make the spatial filter
    int filterSizeT = (int)(spatialSigma * 6 + 1) | 1;
    int filterSizeX = (int)(spatialSigma * 6 + 1) | 1;
    int filterSizeY = (int)(spatialSigma * 6 + 1) | 1;

    // common case optimization
    if (im.frames == 1) filterSizeT = 1;

    Image filter(filterSizeT, filterSizeX, filterSizeY, 1);

    for (int t = 0; t < filter.frames; t++) {
	for (int y = 0; y < filter.height; y++) {
	    for (int x = 0; x < filter.width; x++) {		
		float dt = (t - filter.frames / 2) / spatialSigma;
		float dx = (x - filter.width / 2) / spatialSigma;
		float dy = (y - filter.height / 2) / spatialSigma;
		float distance = dt * dt + dx * dx + dy * dy;
		float value = expf(-distance/2);
		filter(t, x, y)[0] = value;
	    }
	}
    }

    // convolve
    int xoff = (filter.width - 1)/2;
    int yoff = (filter.height - 1)/2;
    int toff = (filter.frames - 1)/2;

    float colorSigmaMult = 1.0f/colorSigma;

    for (int t = 0; t < im.frames; t++) {
        for (int y = 0; y < im.height; y++) {
            for (int x = 0; x < im.width; x++) {
                float totalWeight = 0;
                for (int dt = -toff; dt <= toff; dt++) {		    
                    int imt = t + dt;
		    if (imt < 0) continue;
		    if (imt >= im.frames) break;
                    int filtert = dt + toff;
                    for (int dy = -yoff; dy <= yoff; dy++) {
			int imy = y + dy;
			if (imy < 0) continue;
			if (imy >= im.height) break;
                        int filtery = dy + yoff;
                        for (int dx = -xoff; dx <= xoff; dx++) {
			    int imx = x + dx;
			    if (imx < 0) continue;
			    if (imx >= im.width) break;
                            int filterx = dx + xoff;
                            float weight = filter(filtert, filterx, filtery)[0];
                            float colorDistance = 0;                           
                            for (int c = 0; c < ref.channels; c++) {
                                float diff = (ref(imt, imx, imy)[c] - ref(t, x, y)[c]) * colorSigmaMult;
                                colorDistance += diff * diff;
                            }
                            weight *= fastexp(-colorDistance/2);
                            totalWeight += weight;
                            for (int c = 0; c < im.channels; c++) {
                                out(t, x, y)[c] += weight * im(imt, imx, imy)[c];
                            }
                        }
                    }
                }
                for (int c = 0; c < im.channels; c++) {
                    out(t, x, y)[c] /= totalWeight;
                }
            }
        }
    }
    

    return out;
}



void BilateralSharpen::help() {
    printf("\n-bilateralsharpen sharpens using a bilateral filter to avoid ringing. The\n"
           "three arguments are the spatial and color standard deviations, and the sharpness.\n"
           "A sharpness of zero has no effect; a sharpness of 1 is significant.\n\n"
           "Usage: ImageStack -load input.jpg -bilateralsharpen 1.0 0.2 1 -save sharp.jpg\n\n");
}

void BilateralSharpen::parse(vector<string> args) {
    assert(args.size() == 3, "-bilateralsharpen takes three arguments");
    Image im = apply(stack(0), readFloat(args[0]), readFloat(args[1]), readFloat(args[2]));
    pop();
    push(im);
}

Image BilateralSharpen::apply(Window im, float spatialSigma, float colorSigma, float sharpness) {
    Image blurred = Bilateral::apply(im, spatialSigma, colorSigma);
    Image out(im);

    float *blurPtr = blurred(0, 0, 0);
    float *outPtr = out(0, 0, 0);
    float blurWeight = -sharpness, outWeight = 1+sharpness;
    for (int i = 0; i < im.frames * im.width * im.height * im.channels; i++) {
	*outPtr = (*outPtr) * outWeight + (*blurPtr) * blurWeight;
	outPtr++;
	blurPtr++;
    }

    return out;
}

					 
void ChromaBlur::help() {
	printf("\n-chromablur blurs an image in the chrominance channels only. It is a good way\n"
           "of getting rid of chroma noise without apparently blurring the image. The two\n"
           "arguments are the standard deviation in space and color of the bilateral filter.\n" 
           "Usage: ImageStack -load input.jpg -chromablur 2 -save output.png\n\n");
}

void ChromaBlur::parse(vector<string> args) {
    assert(args.size() == 2, "-chromablur takes two arguments");
    Image im = apply(stack(0), readFloat(args[0]), readFloat(args[1]));
    pop();
    push(im);
}

Image ChromaBlur::apply(Window im, float spatialSigma, float colorSigma) {
    assert(im.channels == 3, "input must be a rgb image\n");
    Image yuv = ColorConvert::rgb2yuv(im);
    Image luminance = ColorConvert::rgb2y(im);

    // blur chrominance
    yuv = Bilateral::apply(yuv, spatialSigma, colorSigma);

    // restore luminance
    for (int t = 0; t < im.frames; t++) {
        for (int y = 0; y < im.height; y++) {
            for (int x = 0; x < im.width; x++) {
                yuv(t, x, y)[0] = luminance(t, x, y)[0];
            }
        }
    }

    return ColorConvert::yuv2rgb(yuv);
}



void MedianFilter::help() {
    printf("-medianfilter applies a median filter with a circular support. The sole argument is\n"
           "the pixel radius of the filter.\n\n"
           "Usage: ImageStack -load input.jpg -median 10 -save output.jpg\n\n");
}

void MedianFilter::parse(vector<string> args) {
    assert(args.size() == 1, "-medianfilter takes one argument\n");
    int radius = readInt(args[0]);
    assert(radius > -1, "radius must be positive");
    Image im = apply(stack(0), radius);
    pop();
    push(im);
}

Image MedianFilter::apply(Window im, int radius) {
    return PercentileFilter::apply(im, radius, 0.5);
}

void PercentileFilter::help() {
    printf("-percentilefilter selects a given statistical percentile over a circular support\n"
           "around each pixel. The two arguments are the support radius, and the percentile.\n"
           "A percentile argument of 0.5 gives a median filter, whereas 0 or 1 give min or\n"
           "max filters.\n\n"
           "Usage: ImageStack -load input.jpg -percentilefilter 10 0.25 -save dark.jpg\n\n");
}

void PercentileFilter::parse(vector<string> args) {
    assert(args.size() == 2, "-percentilefilter takes two arguments\n");
    int radius = readInt(args[0]);
    float percentile = readFloat(args[1]);
    assert(0 <= percentile && percentile <= 1, "percentile must be between zero and one");
    if (percentile == 1) percentile = 0.999;
    assert(radius > -1, "radius must be positive");
    Image im = apply(stack(0), radius, percentile);
    pop();
    push(im);
}

Image PercentileFilter::apply(Window im, int radius, float percentile) {
    // make the histogram
    const int buckets = 256;

    vector<int> histogram(buckets);

    Image out(im.frames, im.width, im.height, im.channels);

    // make the filter edge profile
    vector<int> edge(radius*2+1);

    for (int i = 0; i < 2*radius+1; i++) {
        edge[i] = (int)(sqrtf(radius*radius - (i - radius)*(i-radius)) + 0.0001f);
    }
    
    for (int c = 0; c < im.channels; c++) {
        for (int t = 0; t < im.frames; t++) {
            for (int y = 0; y < im.height; y++) {
                // initialize the histogram for this scanline
                for (int i = 0; i < buckets; i++) histogram[i] = 0;
                int lteq = 0;
                int total = 0;
                int medianBucket = buckets/2;

                for (int i = 0; i < 2*radius+1; i++) {
                    int xoff = edge[i];
                    int yoff = i - radius;

                    if (y + yoff >= im.height) break;
                    if (y + yoff < 0) continue;

                    for (int j = 0; j <= xoff; j++) {
                        if (j >= im.width) break;
                        float val = im(t, j, y+yoff)[c];
                        int bucket = HDRtoLDR(val);
                        histogram[bucket]++;
                        if (bucket <= medianBucket) lteq++;
                        total++;
                    }                    
                }
            
                for (int x = 0; x < im.width; x++) {
                    // adjust the median bucket downwards
                    while (lteq > total * percentile && medianBucket > 0) {
                        lteq -= histogram[medianBucket];
                        medianBucket--;
                    }
                    
                    // adjust the median bucket upwards
                    while (lteq <= total * percentile && medianBucket < buckets-1) {
                        medianBucket++;
                        lteq += histogram[medianBucket];
                    }
                    
                    out(t, x, y)[c] = LDRtoHDR(medianBucket);
                    
                    // move the histogram to the right
                    for (int i = 0; i < radius*2+1; i++) {
                        int xoff = edge[i];
                        int yoff = i - radius;
                        
                        if (y + yoff >= im.height) break;
                        if (y + yoff < 0) continue;
                        
                        // subtract old value
                        if (x - xoff >= 0) {
                            float val = im(t, x-xoff, y+yoff)[c];
                            int bucket = HDRtoLDR(val);
                            histogram[bucket]--;
                            if (bucket <= medianBucket) lteq--;
                            total--;
                        }
                        
                        // add new value
                        if (x + xoff + 1 < im.width) {
                            float val = im(t, x+xoff+1, y+yoff)[c];
                            int bucket = HDRtoLDR(val);
                            histogram[bucket]++;
                            if (bucket <= medianBucket) lteq++;
                            total++;
                        }
                    }
                }                
            }
        }
    }

    return out;
}



void CircularFilter::help() {
    printf("\n-circularfilter convolves the image with a uniform circular kernel. It is a good\n"
           "approximate to out of focus blur. The sole argument is the radius of the filter.\n\n"
           "Usage: ImageStack -load in.jpg -circularfilter 10 -save out.jpg\n\n");
}

void CircularFilter::parse(vector<string> args) {
    assert(args.size() == 1, "-circularfilter takes one argument\n");

    Image im = apply(stack(0), readInt(args[0]));
    pop();
    push(im);
}

Image CircularFilter::apply(Window im, int radius) {
    Image out(im.frames, im.width, im.height, im.channels);

    // maintain the average response currently under the filter, and the number of pixels under the filter
    float average = 0;
    int count = 0;

    // make the filter edge profile
    vector<int> edge(radius*2+1);
    for (int i = 0; i < 2*radius+1; i++) {
        edge[i] = (int)(sqrtf(radius*radius - (i - radius)*(i-radius)) + 0.0001f);
    }

    // figure out the filter area
    for (int i = 0; i < 2*radius+1; i++) {
        count += edge[i]*2+1;
    }

    float invArea = 1.0f/count;

    for (int c = 0; c < im.channels; c++) {
        for (int t = 0; t < im.frames; t++) {
            for (int y = 0; y < im.height; y++) {
                average = 0;
                // initialize the average and count
                for (int i = 0; i < 2*radius+1; i++) {
                    int xoff = edge[i];
                    int yoff = i - radius;
                    int realY = MIN(MAX(y + yoff, 0), im.height-1);

                    for (int x = -xoff; x <= xoff; x++) {
                        int realX = MIN(MAX(x, 0), im.width-1);
                        float val = im(t, realX, realY)[c];
                        average += val;
                    }                    
                }
            
                for (int x = 0; x < im.width; x++) {
                    out(t, x, y)[c] = average * invArea;
                    
                    // move the histogram to the right
                    for (int i = 0; i < radius*2+1; i++) {                        
                        int realXOld = MAX(0, x-edge[i]);
                        int realXNew = MIN(x+edge[i]+1, im.width-1);
                        int realY = MIN(MAX(0, y+i-radius), im.height-1);
                        
                        // add new value, subtract old value
                        average += im(t, realXNew, realY)[c];
                        average -= im(t, realXOld, realY)[c];
                    }
                }                
            }
        }
    }

    return out;
}



